import os
from logging.config import fileConfig

from sqlalchemy import engine_from_config
from sqlalchemy import pool, text

from alembic import context

from getitfixed.scripts import wait_for_db

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(context.config.config_file_name)

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_config():
    conf = context.config.get_section(context.config.config_ini_section)

    # Load config from c2cgeoportal if available
    app_cfg = context.config.get_main_option("app.cfg")
    if app_cfg:
        from c2c.template.config import config

        config.init(context.config.get_main_option("app.cfg"))
        conf.update(config.get_config())

    if "sqlalchemy.url" not in conf:
        conf[
            "sqlalchemy.url"
        ] = "postgresql://{PGUSER}:{PGPASSWORD}@{PGHOST}:{PGPORT}/{PGDATABASE}".format(
            **os.environ
        )
    conf.update(
        {"version_table_schema": conf.get("getitfixed", {}).get("schema", "getitfixed")}
    )

    # for 'autogenerate' support
    from getitfixed import models  # noqa

    conf["target_metadata"] = models.meta.Base.metadata

    return conf


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    conf = get_config()
    context.configure(url=conf["sqlalchemy.url"], literal_binds=True, **conf)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    conf = get_config()
    schema = conf.get("getitfixed", {}).get("schema", "getitfixed")

    connectable = engine_from_config(
        conf, prefix="sqlalchemy.", poolclass=pool.NullPool
    )

    def include_object(
        obj, name, type_, reflected, compare_to
    ):  # pylint: disable=unused-argument
        if type_ == "table":
            return obj.schema == schema
        else:
            return obj.table.schema == schema

    wait_for_db(connectable)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            include_schemas=True,
            include_object=include_object,
            **conf
        )

        connection.execute(text('CREATE SCHEMA IF NOT EXISTS "{}";'.format(schema)))

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""First migration

Revision ID: 0b12402a098e
Revises:
Create Date: 2020-05-20 16:47:03.128539

"""
from alembic import op
import sqlalchemy as sa
import geoalchemy2


# revision identifiers, used by Alembic.
revision = "0b12402a098e"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "category",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("label_fr", sa.String(length=50), nullable=True),
        sa.Column("label_en", sa.String(length=50), nullable=True),
        sa.Column("email", sa.String(length=50), nullable=False),
        sa.Column("icon", sa.String(length=150), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_category")),
        schema="getitfixed",
    )
    op.create_table(
        "type",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("label_fr", sa.String(length=50), nullable=True),
        sa.Column("label_en", sa.String(length=50), nullable=True),
        sa.Column("category_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["category_id"],
            ["getitfixed.category.id"],
            name=op.f("fk_type_category_id_category"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_type")),
        schema="getitfixed",
    )
    op.create_table(
        "issue",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("hash", sa.Text(), nullable=False),
        sa.Column(
            "request_date", sa.Date(), server_default=sa.text("now()"), nullable=False
        ),
        sa.Column("type_id", sa.Integer(), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "new",
                "validated",
                "in_progress",
                "waiting_for_reporter",
                "resolved",
                name="status",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("localisation", sa.String(length=254), nullable=False),
        sa.Column(
            "geometry",
            geoalchemy2.types.Geometry(
                geometry_type="POINT", srid=4326, management=True
            ),
            nullable=True,
        ),
        sa.Column("firstname", sa.String(length=100), nullable=False),
        sa.Column("lastname", sa.String(length=100), nullable=False),
        sa.Column("phone", sa.String(length=20), nullable=False),
        sa.Column("email", sa.String(length=100), nullable=False),
        sa.ForeignKeyConstraint(
            ["type_id"], ["getitfixed.type.id"], name=op.f("fk_issue_type_id_type")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_issue")),
        sa.UniqueConstraint("hash", name=op.f("uq_issue_hash")),
        schema="getitfixed",
    )
    op.create_table(
        "event",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("issue_id", sa.Integer(), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "new",
                "validated",
                "in_progress",
                "waiting_for_reporter",
                "resolved",
                name="status",
                native_enum=False,
            ),
            nullable=False,
        ),
        sa.Column(
            "date",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("comment", sa.Text(), nullable=True),
        sa.Column("private", sa.Boolean(), nullable=True),
        sa.Column(
            "author",
            sa.Enum("customer", "admin", name="author", native_enum=False),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["issue_id"], ["getitfixed.issue.id"], name=op.f("fk_event_issue_id_issue")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_event")),
        schema="getitfixed",
    )
    op.create_table(
        "photo",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("filename", sa.Text(), nullable=True),
        sa.Column("data", sa.LargeBinary(), nullable=False),
        sa.Column("issue_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["issue_id"], ["getitfixed.issue.id"], name=op.f("fk_photo_issue_id_issue")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_photo")),
        schema="getitfixed",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("photo", schema="getitfixed")
    op.drop_table("event", schema="getitfixed")
    op.drop_table("issue", schema="getitfixed")
    op.drop_table("type", schema="getitfixed")
    op.drop_table("category", schema="getitfixed")
    # ### end Alembic commands ###



"""Add field issue.private

Revision ID: 6d8c64186efa
Revises: f12edd1058e7
Create Date: 2020-06-04 07:47:52.848957

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "6d8c64186efa"
down_revision = "f12edd1058e7"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "issue",
        sa.Column(
            "private", sa.Boolean(), nullable=False, server_default=sa.text("False")
        ),
        schema="getitfixed",
    )


def downgrade():
    op.drop_column("issue", "private", schema="getitfixed")



import pytest

from getitfixed.models.getitfixed import Category, Type, Issue, STATUS_IN_PROGRESS


ICONS = [
    ("https://server.com/static/image.png",),
    ("/static/image.png",),
    ("static/image.png",),
    ("static://getitfixed:static/image.png", "http://localhost/getitfixed_static/image.png"),
    (None, "http://localhost/getitfixed_static/icons/cat-default.png"),
]


def category(**kwargs):
    defaults = {
        "label_en": "Category",
        "label_fr": "Catégorie",
        "email": "is@abit.ch",
    }
    return Category(**{**defaults, **kwargs})


def type_(**kwargs):
    defaults = {
        "label_en": "Type (en)",
        "label_fr": "Type (fr)",
    }
    return Type(
        category=category(**kwargs.pop("category", {})),
        **{**defaults,**kwargs},
    )


def issue(**kwargs):
    defaults = {
        "description": "description",
        "localisation": "localisation",
        "firstname": "firstname",
        "lastname": "lastname",
        "phone": "0479000000",
        "status": STATUS_IN_PROGRESS,
        "email": "firstname.lastname@domain.net",
    }
    return Issue(
        type=type_(**kwargs.pop("type", {})),
        **{**defaults, **kwargs},
    )


class TestCategory():
    def test_label(self):
        c = category()
        assert c.label("en") == "Category"
        assert c.label("fr") == "Catégorie"

    @pytest.mark.parametrize("icon,icon_url", [(i[0], i[-1]) for i in ICONS])
    @pytest.mark.usefixtures("app_env")
    def test_icon_url(self, app_env, icon, icon_url):
        request = app_env["request"]
        assert category(icon=icon).icon_url(request) == icon_url


class TestType():
    def test_label(self):
        t = type_()
        assert t.label("en") == "Type (en)"
        assert t.label("fr") == "Type (fr)"


class TestIssue:
    def test_status_i18n(self):
        i = issue()
        assert i.status_de == "In progress"  # FIXME: after de translations are done
        assert i.status_en == "In progress"
        assert i.status_fr == "En cours"

    @pytest.mark.parametrize("icon,icon_url", [(i[0], i[-1]) for i in ICONS])
    @pytest.mark.usefixtures("app_env")
    def test_icon_url(self, app_env, icon, icon_url):
        request = app_env["request"]
        i = issue(
            type={
                "category": {"icon": icon}
            }
        )
        assert i.category.icon == icon
        assert i.icon_url(request) == icon_url



import pytest
import webtest

from getitfixed.models.getitfixed import Category, Type
from getitfixed.scripts.setup_test_data import WMS_LAYERS

from ..model_test import ICONS


@pytest.fixture(scope="function")
@pytest.mark.usefixtures("dbsession", "transact")
def categories_test_data(dbsession, transact):
    del transact

    categories = []
    for i in range(5):
        categories.append(
            Category(
                label_en="Category «{}»".format(i),
                label_fr="Catégorie «{}»".format(i),
                email="{}.is@abit.ch".format(i),
                icon=ICONS[i][0],
            )
        )
    dbsession.add_all(categories)

    types = []
    for i in range(15):
        types.append(
            Type(
                label_en="Type «{}»".format(i),
                label_fr="Type «{}»".format(i),
                category=categories[i % 3],
                wms_layer=WMS_LAYERS[i % 4],
            )
        )
    dbsession.add_all(types)

    yield {"categories": categories, "types": types}


@pytest.mark.usefixtures("categories_test_data", "test_app")
class TestCategories():

    @pytest.mark.parametrize("locale", ["en", "fr"])
    def test_success(self, test_app, categories_test_data, locale):
        categories = categories_test_data["categories"]

        test_app.set_cookie("_LOCALE_", locale)
        result = test_app.get("/getitfixed/categories.json", status=200).json

        assert len(result) == len(categories)
        for c, exp_c, icon in zip(result, categories, ICONS):
            assert c["id"] == exp_c.id
            assert c["label"] == exp_c.label(locale)
            assert c["icon"] == icon[-1]
            for t, exp_t in zip(c["types"], exp_c.types):
                assert t["id"] == exp_t.id
                assert t["label"] == exp_t.label(locale)
                assert t["wms_layer"] == exp_t.wms_layer



from pyramid.view import view_config, view_defaults

from getitfixed.views.admin.photos import PhotoViews


@view_defaults(match_param=("application=getitfixed_private", "table=photos"))
class PhotoViews(PhotoViews):
    @view_config(route_name="c2cgeoform_item", request_method="GET")
    def get(self):
        return super().get()



"""Add field type.wms_layer

Revision ID: f12edd1058e7
Revises: 0b12402a098e
Create Date: 2020-05-20 16:49:55.269463

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "f12edd1058e7"
down_revision = "0b12402a098e"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "type",
        sa.Column("wms_layer", sa.String(length=255), nullable=True),
        schema="getitfixed",
    )


def downgrade():
    op.drop_column("type", "wms_layer", schema="getitfixed")



from pyramid.view import view_config

from getitfixed.models.getitfixed import Category


@view_config(
    route_name="c2cgeoform_index",
    match_param=("application=getitfixed", "table=categories.json"),
    request_method="GET",
    renderer="json",
)
def categories(request):
    return [
        {
            "id": c.id,
            "label": c.label(request.locale_name),
            "icon": c.icon_url(request),
            "types": [
                {
                    "id": t.id,
                    "label": t.label(request.locale_name),
                    "wms_layer": t.wms_layer,
                }
                for t in c.types
            ],
        }
        for c in request.dbsession.query(Category)
    ]



"""Add field photo.hash

Revision ID: 73e05f3650a7
Revises: 6d8c64186efa
Create Date: 2020-07-24 09:05:07.981495

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "73e05f3650a7"
down_revision = "6d8c64186efa"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "photo",
        sa.Column("hash", sa.Text(), nullable=True, unique=True),
        schema="getitfixed",
    )
    op.execute(
        sa.text(
            "UPDATE getitfixed.photo"
            " SET hash = uuid_in(md5(random()::text || clock_timestamp()::text)::cstring);"
        )
    )
    op.alter_column("photo", "hash", nullable=False, schema="getitfixed")


def downgrade():
    op.drop_column("photo", "hash", schema="getitfixed")
